/*
 * VIA VT86C100A Rhine-II and VIA VT3043 Rhine Based Card Add-on
 * for the BeOS Release 5
 */
 
/*
 * Common Methods for Ethernet Drivers
 */
#include "via-rhine-device.h"
#include "via-rhine-netdebug.h"
#include <string.h>
#include <errno.h>
#include <new>
#include <BeBuild.h>	// for _PR2_COMPATIBLE_

//#define __VIARHINE_DEBUG__

#if _PR2_COMPATIBLE_

int compat_open(const char *dev);

#else

#define compat_open(dev) open(dev, O_RDWR, 0)

#endif

#ifdef __VIARHINE_DEBUG__
int debug_printf(char *format, ...)
{
	va_list vl;
	char    c[2048];
	int     i;
	static int f = 0;
	
	va_start(vl, format);
	i = vsprintf(c, format, vl);
	va_end(vl);

	if (f == 0)
		f = open("/boot/via-rhine-addon.log", O_WRONLY | O_APPEND | O_TEXT);

	write(f, c, strlen(c));
	return i;
}
#endif

int toss(int fd)
{
	char junk[1536];

#ifdef __VIARHINE_DEBUG__	
	debug_printf("toss\n");
#endif

	return (read(fd, junk, sizeof(junk)));
}

EtherDevice::EtherDevice()
{
	fd          = -1;
	nonblock    = 0;
	got_address = 0;
	framesize   = 0;
}

BIpDevice *EtherDevice::OpenIP(void)
{
	return NULL;
}

BNetPacket *EtherDevice::AllocPacket(void)
{
	return new BStandardPacket();
}

void EtherDevice::SendPacket(BNetPacket *packet)
{
	write(fd, packet->Data(), packet->Size());
	delete packet;
}

status_t EtherDevice::AddMulticastAddress(const char *address)
{
#ifdef __VIARHINE_DEBUG__
	debug_printf("EtherDevice::AddMulticastAddress: %02x:%02x:%02x:%02x:%02x:%02x\n",
			(unsigned char)address[0],
			(unsigned char)address[1],
			(unsigned char)address[2],
			(unsigned char)address[3],
			(unsigned char)address[4],
			(unsigned char)address[5]);
#endif

	return ioctl(fd, ETHER_ADDMULTI, address);
}

status_t EtherDevice::RemoveMulticastAddress(const char *address)
{
#ifdef __VIARHINE_DEBUG__
	debug_printf("EtherDevice::RemoveMulticastAddress: %02x:%02x:%02x:%02x:%02x:%02x\n",
			(unsigned char)address[0],
			(unsigned char)address[1],
			(unsigned char)address[2],
			(unsigned char)address[3],
			(unsigned char)address[4],
			(unsigned char)address[5]);
#endif

	return ioctl(fd, ETHER_REMMULTI, address);
}

void EtherDevice::Address(char *address)
{
	if (!got_address)
	{
		ioctl(fd, ETHER_GETADDR, this->address);
		got_address = true;

#ifdef __VIARHINE_DEBUG__
		debug_printf("EtherDevice::Address: %02x:%02x:%02x:%02x:%02x:%02x\n",
				(unsigned char)this->address[0],
				(unsigned char)this->address[1],
				(unsigned char)this->address[2],
				(unsigned char)this->address[3],
				(unsigned char)this->address[4],
				(unsigned char)this->address[5]);
#endif
	}

	memcpy(address, this->address, sizeof(this->address));
}

status_t EtherDevice::SetPromiscuous(bool yes)
{
	int promisc = yes;

#ifdef __VIARHINE_DEBUG__
	debug_printf("EtherDevice::SetPromiscuous (%u)\n", (unsigned int)yes);
#endif
	
	return ioctl(fd, ETHER_SETPROMISC, &promisc);
}

int EtherDevice::Init(const char *devicelink, char *params)
{
	fd = compat_open(devicelink);

	if (fd < 0)
	{
		return (-1);
	}

	if (ioctl(fd, ETHER_INIT, params, 0) < 0)
	{
#ifdef __VIARHINE_DEBUG__
		debug_printf("EtherDevice::Init Failed\n");
#endif
		close(fd);
		fd = -1;
		return (-1);
	}
	
#ifdef __VIARHINE_DEBUG__
	debug_printf("EtherDevice::Init Success!\n");
#endif
	return (fd);
}

net_device_type EtherDevice::Type(void)
{
	return B_ETHER_NET_DEVICE;
}

void EtherDevice::Statistics(FILE *f)
{
}

unsigned int EtherDevice::MaxPacketSize(void)
{
	if (framesize == 0)
	{
		if (ioctl(fd, ETHER_GETFRAMESIZE, &framesize) < 0)
		{
			framesize = 1536;
		}
#ifdef __VIARHINE_DEBUG__
		debug_printf("EtherDevice::MaxPacketSize = %u\n", framesize);
#endif
	}

	return framesize;
}

int EtherDevice::SetMaxPacketSize(unsigned int framesize)
{
#ifdef __VIARHINE_DEBUG__
	debug_printf("EtherDevice::SetMaxPacketSize(%u)\n", framesize);
#endif

	if (framesize > 1536)
	{
		return B_ERROR;
	}

	this->framesize = framesize;
	return B_NO_ERROR;
}

void EtherDevice::Close(void)
{
	if (fd >= 0)
	{
		close(fd);
		fd = -1;
	}
}

BNetPacket *EtherDevice::ReceivePacket(void)
{
	BNetPacket *packet;

	if (nonblock)
	{
		nonblock = 0;
		ioctl(fd, ETHER_NONBLOCK, &nonblock, sizeof(nonblock));
	}

	/*
	 * Wait for a packet
	 */
	do
	{
		packet = ReadPacket();     /* NULL if out of memory */
	} while (packet == NULL); 

	/*
	 * Got a packet
	 */
	return packet;
}

BNetPacket *EtherDevice::ReadPacket(void)
{
	BNetPacket *packet;
	int         len;

	packet = NULL;
	try
	{
		/*
		 * Normal case: assume we don't run out of memory
		 */
		packet = AllocPacket();
		packet->SetSize(2048);
		len = read(fd, packet->Data(), packet->Size());
		
		if (len <= 0)
		{
#ifdef __VIARHINE_DEBUG__
			debug_printf("EtherDevice::read_packet error (%u)\n", len);
#endif
			delete packet;
			return NULL;
		}

#ifdef __VIARHINE_DEBUG__
		debug_printf("EtherDevice::read_packet (%u)\n", len);
#endif
		packet->SetSize(len);
	}
	catch (bad_alloc)
	{
#ifdef __VIARHINE_DEBUG__
		debug_printf("EtherDevice::read_packet: bad_alloc throwed\n");
#endif

		if (!nonblock)
		{
			/*
			 * If we are blocking, must do a blocking read
			 */
			if (toss(fd) <= 0)
			{
				return NULL;
			}
		}
		if (packet != NULL)
		{
			/*
			 * A packet was allocated, but SetSize failed
			 */
			delete packet;
		}
		
		return NULL;
	}

	return packet;
}

#if _PR2_COMPATIBLE_
int compat_open(const char *dev)
{
	if (strcmp(dev, "/dev/via-rhine") == 0)
	{
		return (open("/dev/net/via-rhine", O_RDWR, 0));
	}
	if (strcmp(dev, "/dev/ether") == 0)
	{
		return (open("/dev/net/ether", O_RDWR, 0));
	}
	
	return (open(dev, O_RDWR, 0));
}
#endif /* _PR2_COMPATIBLE_ */
